from .decorators import chat_completion_raw_response_decorator
from .decorators import chat_completion_decorator
from .decorators import client_decorator


def track_openai():
    """
    The function `track_openai` decorates the OpenAI client to add
    additional functionalities.
    """
    
    try:
        
        import openai
        
        # Force OpenAI to track the tokens usage separately for each key
        original_init = openai.OpenAI.__init__
        
        def new_init(self, *args, **kwargs):
        
            original_init(self, *args, **kwargs)
            
            # Raw chat completion
            self.chat.completions.with_raw_response.create = chat_completion_raw_response_decorator(self.chat.completions.with_raw_response.create)
            
            # With options
            # No need to decorate it, it is already decorated by the below decorator
            
            # Basic chat completion
            self.chat.completions.create = chat_completion_decorator(self.chat.completions.create)
        
        openai.OpenAI.__init__ = new_init
        
        # Force parameter organization to be specified, either as argument or as env variable
        openai.OpenAI.__init__ = client_decorator(openai.OpenAI.__init__)
        
    except ImportError:

        print("OpenAI not installed. Please install it with 'pip install openai'")
        


import tiktoken


def num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613"):
    """Return the number of tokens used by a list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model in {
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
        }:
        tokens_per_message = 3
        tokens_per_name = 1
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-3.5-turbo" in model:
        print("Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613.")
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
    elif "gpt-4" in model:
        print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
        return num_tokens_from_messages(messages, model="gpt-4-0613")
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens

    
def num_tokens_from_string(string:str, model:str) -> int:
    """
    Returns the number of tokens in a text string
    """
    try:
        encoding = tiktoken.encoding_for_model(model).name
    except:
        encoding = 'cl100k_base'
    try:
        encoding = tiktoken.get_encoding(encoding)
        num_tokens = len(encoding.encode(string))
        return num_tokens
    except:
        return 0



from bs4 import BeautifulSoup
import requests
import json
import os


def scrape_openai_price() -> dict:
    """
    Thi function scrapes and return a dict of OpenAI's pricing,
    where every key is a model name.
    """
    
    def clean_txt(txt):
        return txt.encode("ISO-8859-1").decode("UTF-8").replace("\xa0", "").replace("Â·", "")

    response = requests.get("https://openai.com/pricing")

    soup = BeautifulSoup(response.text, 'html.parser')

    container_div = soup.find("div", {"id": "content"})
    models_div = container_div.find_all("div", {"class": "ui-block ui-block--pricing-table"})

    prices = {}

    for div in models_div:
        
        #father_model_name = div.get("id")
        #prices[father_model_name] = {}
        
        # find the table in div
        table = div.find("table")
        # find all tr in table
        trs = table.find_all("tr")
        
        # Header row
        columns = [td.text for td in trs[0].find_all("td")]
        
        # Content rows
        for row in trs[1:]:
            
            # Cells
            cells = [td for td in row.find_all("td")]
            assert len(cells) == len(columns)
            
            # First cell
            submodel_name = cells[0].find('span')
            
            if submodel_name:
            
                submodel_name = clean_txt(submodel_name.text)
                #prices[father_model_name][submodel_name] = {}
                prices[submodel_name] = {}
                
                # Other cells
                for i, (cell, column) in enumerate(zip(cells[1:], columns[1:])):
                    
                    # Extract content from cell
                    spans = cell.find_all("span")
                    if spans:
                        if len(spans) > 1:
                            text = "".join([clean_txt(span.text) for span in spans])
                            
                        else:
                            text = clean_txt(spans[0].text)
                        
                        #prices[father_model_name][submodel_name][column] = text
                        prices[submodel_name][column] = text
        
    return prices

def parse_dict(prices:dict) -> dict:
    """
    This function takes the dict of prices and returns a new dict
    parsed and cleaned to contain floats and integers instead of strings,
    ready to be multiplied by the number of tokens.
    """
    
    prices2 = {}
    
    for k,v in prices.items():
        
        prices2[k] = {}
        
        if "gpt" in k.lower() or "older models" in k.lower():
        
            for header, full_price in v.items():
                header = header.lower().replace("usage","").strip()
                full_price = full_price.lower()
                price = float(full_price.split("/")[0].replace("$", "").strip())
                quotient = int(full_price.split("/")[1].replace("tokensm","").replace("tokensk","").replace("tokens","").replace("k","000").replace("m","000000").strip())
                prices2[k][header] = {"price": price, "quotient": quotient, "currency": "$", "um": "tokens"}
        else:
            prices2[k] = v
            
    return prices2
    
def update_pricing(verbose:bool=False) -> dict:
    """
    Main function to update the pricing and save it to a json file.
    :param verbose: if True, prints the prices
    """
    prices = scrape_openai_price()
    prices = parse_dict(prices)
    with open("prices.json", "w") as f:
        json.dump(prices, f)
    if verbose:
        print(prices)
    return prices

def get_pricing(model:str, prompt_tokens:int, completion_tokens:int) -> tuple:
    """
    Given a model name and the number of tokens used for prompt and completion,
    returns the price for the prompt and the completion.
    :param model: the openai model name
    :param prompt_tokens: the number of tokens used for the prompt
    :param completion_tokens: the number of tokens used for the completion
    :return: a tuple with the price for the prompt and the completion
    """
    
    if not os.path.exists("prices.json"):
        _ = update_pricing()
    with open("prices.json", "r") as f:
        prices = json.load(f)
    
    try:
        model_pricing = prices.get(model,None)
        if not model_pricing:
            # if 'gpt-3.5' in model.lower():
            #     input_price = 0.0005
            #     input_quotient = 1000
            #     output_price = 3*0.0005
            #     output_quotient = 1000
            # elif 'gpt-4' in model.lower():
            #     input_price = 0.0015
            #     input_quotient = 1000
            #     output_price = 3*0.0015
            #     output_quotient = 1000
            # else:
            #    return 0,0
            print(f"Model {model} not found. Price will not be added to log.")
            return 0,0
            
        input_price = model_pricing["input"]["price"]
        input_quotient = model_pricing["input"]["quotient"]
        prompt_price = input_price * (prompt_tokens / input_quotient)
        
        output_price = model_pricing["output"]["price"]
        output_quotient = model_pricing["output"]["quotient"]
        completion_price = output_price * (completion_tokens / output_quotient)
        
        return prompt_price, completion_price
        
    except Exception as e:
        print("Exception in get_pricing():",e)
        return 0,0


if __name__ == "__main__":
    prices = update_pricing()
    print(prices)


#https://github.com/openai/openai-python/issues/715#issuecomment-1809203346

import datetime
from unittest.mock import patch

from openai.types.chat import ChatCompletionMessage
from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.completion_usage import CompletionUsage
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk, ChoiceDelta
from openai.types.chat.chat_completion_chunk import Choice as StreamChoice


def mock_chat_completion(response: str, model: str = "gpt-3.5-turbo") -> ChatCompletion:
    return ChatCompletion(
        id="foo",
        model=model,
        object="chat.completion",
        choices=[
            Choice(
                finish_reason="stop",
                index=0,
                message=ChatCompletionMessage(
                    content=response,
                    role="assistant",
                ),
            )
        ],
        usage=CompletionUsage(
            prompt_tokens=0,
            completion_tokens=10,
            total_tokens=10,
        ),
        created=int(datetime.datetime.now().timestamp()),
    )


def mock_stream_chat_completion(response: str, model: str = "gpt-3.5-turbo"):
    for token in response:
        yield ChatCompletionChunk(
            id="foo",
            model=model,
            object="chat.completion.chunk",
            choices=[
                StreamChoice(
                    index=0,
                    finish_reason=None,
                    delta=ChoiceDelta(
                        content=token,
                        role="assistant",
                    )
                ),
            ],
            created=int(datetime.datetime.now().timestamp()),
        )


@patch("openai.resources.chat.Completions.create")
def test_chat_completion(openai_create):
    from openai import OpenAI
    EXPECTED_RESPONSE = "The mock is working! ;)"
    openai_create.return_value = mock_chat_completion(EXPECTED_RESPONSE)
    client = OpenAI(api_key="sk-...")
    r = client.chat.completions.create(
        messages=[{"role": "user", "content": "Do you know any jokes?"}],
        model="gpt-3.5-turbo",
    )
    response = r.choices[0].message.content
    assert response == EXPECTED_RESPONSE


@patch("openai.resources.chat.Completions.create")
def test_stream_chat_completion(openai_create):
    print("aaa  ")
    from openai import OpenAI
    EXPECTED_RESPONSE = "The mock is STILL working! ;)"
    openai_create.return_value = mock_stream_chat_completion(EXPECTED_RESPONSE)
    client = OpenAI(api_key="sk-...")
    stream = client.chat.completions.create(
        messages=[{"role": "user", "content": "Do you know any jokes?"}],
        model="gpt-3.5-turbo",
        stream=True,
    )

    response = ""
    chunk_count = 0
    for chunk in stream:
        response += (chunk.choices[0].delta.content or "")
        print("Chunk:",chunk)
        chunk_count += 1
    assert response == EXPECTED_RESPONSE
    assert chunk_count == len(EXPECTED_RESPONSE)
    
test_stream_chat_completion()


import os
import setuptools

with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'README.md'), 'r') as f:
    long_description = f.read()

setuptools.setup(
    name = "apikeylogger",
    version = "1.1.1",
    author = "Federico Romeo",
    author_email = "federico.romeo.98@gmail.com",
    description = "This library allows you to log the OpenAI api usage *by key* without having to change your code",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/federicoromeo/apikeylogger",
    keywords  =  ['openai', 'apikey', 'api', 'logger', 'tracker'],
    project_urls = {'Source': 'https://github.com/federicoromeo/apikeylogger'},
    license = 'MIT',
    packages = setuptools.find_packages(),
    include_package_data = True,
    classifiers = [],
    python_requires = ">=3.6",
    install_requires = [
        "openai >= 1.6.0",
        "python-dotenv >= 0.18.0",
        "tiktoken",
        "beautifulsoup4"
    ],
)



import os
import json
from .pricing import get_pricing
from .tiktoken_count import num_tokens_from_string, num_tokens_from_messages


def client_decorator(func):
    """
    Decorator that forces the client to have the organization in kwargs
    
    :param func: The function to decorate
    
    :return: The decorated function
    """
    
    def wrapper(*args, **kwargs):
        
        #logging.warning(" Decorating the client")

        if 'organization' not in kwargs.keys():
            if not 'OPENAI_ORG_ID' in os.environ.keys():
                raise Exception("Please specify the 'organization' when instantiating the OpenAI client, or insert it in the environment variable as 'OPENAI_ORG_ID'. Find yours at 'https://platform.openai.com/account/organization'")
        
        response = func(*args, **kwargs)
        
        return response
    
    return wrapper


def chat_completion_raw_response_decorator(func):
    """
    Decorator that logs the tokens used for the request and the response
    for the raw response endpoint
    
    :param func: The function to decorate
    
    :return: The decorated function
    """
    
    def wrapper(*args, **kwargs):
        
        #logging.warning(" Decorating the chat completion function for the raw response")

        def get_model(args, kwargs):
            if 'model' in kwargs.keys():
                return kwargs['model']
            elif len(args) > 0:
                # The model is the second argument
                return args[1]
            else:
                raise Exception("No model specified")
        
        def get_logger_dict():
            if os.path.exists('apikeylogs.json'):
                with open('apikeylogs.json', 'r') as f:
                    d = json.load(f)
                return d
            return {}
        
        def log_tokens(model, prompt_tokens, completion_tokens, prompt_price, completion_price):

            log_dict = get_logger_dict()           

            organization_id = os.environ.get('OPENAI_ORG_ID')
            api_key = os.environ.get('OPENAI_API_KEY')

            # Create the nested dictionary structure
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('prompt_tokens', 0)
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('prompt_price', 0)
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('completion_tokens', 0)
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('completion_price', 0)
            
            # Update token counts and price
            log_dict[organization_id][api_key][model]['prompt_tokens'] += prompt_tokens
            log_dict[organization_id][api_key][model]['prompt_price'] += prompt_price        
            log_dict[organization_id][api_key][model]['completion_tokens'] += completion_tokens
            log_dict[organization_id][api_key][model]['completion_price'] += completion_price

            with open('apikeylogs.json', 'w') as f:
                json.dump(log_dict, f)
                
        response = func(*args, **kwargs)
        
        # Get the tokens from the response
        response_copy = json.loads(response.text)
        prompt_tokens = response_copy['usage']['prompt_tokens']
        completion_tokens = response_copy['usage']['completion_tokens']
        
        # Get the model used for the request
        model = get_model(args, kwargs)
        # Calculate the price based on the used model and tokens
        prompt_price, completion_price = get_pricing(model, prompt_tokens, completion_tokens)
        
        log_tokens(model, prompt_tokens, completion_tokens, prompt_price, completion_price)

        return response
    
    return wrapper


from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai._streaming import Stream

def chat_completion_decorator(func) -> Stream[ChatCompletionChunk]:
    """
    Decorator that logs the tokens used for the request and the response
    
    :param func: The function to decorate
    
    :return: The decorated function
    """
    
    def wrapper(*args, **kwargs) -> Stream[ChatCompletionChunk]:
        
        #logging.warning(" Decorating the chat completion function")

        def get_model(args, kwargs):
            if 'model' in kwargs.keys():
                return kwargs['model']
            elif len(args) > 0:
                # The model is the second argument
                return args[1]
            else:
                raise Exception("No model specified")
        
        def get_logger_dict():
            if os.path.exists('apikeylogs.json'):
                with open('apikeylogs.json', 'r') as f:
                    d = json.load(f)
                return d
            return {}
        
        def log_tokens(model, prompt_tokens, completion_tokens, prompt_price, completion_price):

            log_dict = get_logger_dict()           

            organization_id = os.environ.get('OPENAI_ORG_ID')
            api_key = os.environ.get('OPENAI_API_KEY')

           # Create the nested dictionary structure
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('prompt_tokens', 0)
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('prompt_price', 0)
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('completion_tokens', 0)
            log_dict.setdefault(organization_id, {}).setdefault(api_key, {}).setdefault(model, {}).setdefault('completion_price', 0)
            
            # Update token counts and price
            log_dict[organization_id][api_key][model]['prompt_tokens'] += prompt_tokens
            log_dict[organization_id][api_key][model]['prompt_price'] += prompt_price        
            log_dict[organization_id][api_key][model]['completion_tokens'] += completion_tokens
            log_dict[organization_id][api_key][model]['completion_price'] += completion_price

            with open('apikeylogs.json', 'w') as f:
                json.dump(log_dict, f)
                
        response = func(*args, **kwargs)

        # Get the model used for the request
        model = get_model(args, kwargs)
        
        if "stream" in kwargs.keys() and kwargs["stream"] == True:
            
            if not "messages" in kwargs.keys():
                if len(args) > 0:
                    # The messages are the first argument
                    messages = args[0]
                else:
                    raise Exception("No messages specified in chat completion with stream")
            else:
                messages = kwargs["messages"]
            
            prompt_tokens = num_tokens_from_messages(messages, model)
            
            resp = ""
            chunk_list = []
            for chunk in response:
                chunk_list.append(chunk)
                if chunk.choices[0].delta.content:
                    resp += chunk.choices[0].delta.content
            # Reinsert the response back into the response object
            response.response = chunk_list
            
            completion_tokens = num_tokens_from_string(resp, model)
            
        else:
        
            # Get the tokens from the response
            completion_tokens = response.usage.completion_tokens
            prompt_tokens = response.usage.prompt_tokens
        
        # Calculate the price based on the used model and tokens
        prompt_price, completion_price = get_pricing(model, prompt_tokens, completion_tokens)
        
        log_tokens(model, prompt_tokens, completion_tokens, prompt_price, completion_price)

        return response
    
    return wrapper
